__default_tag_key__ = 'finetuner'

from .fit import fit

# do not change this line manually
# this is managed by git tag and updated on every release
# NOTE: this represents the NEXT release version
__version__ = '0.0.1'
